/*
 *	Copyright 2007  Benjamin C. R. LaHaise, All Rights Reserved.
 *	Permission is hereby granted to copy under the terms of the GPLv2
 *	or later.  See the file LICENSE for details.
 */
#include "config.h"
#include "babd.h"
#include "ctrlfd.h"

#include "bpppoe.h"

#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <stdlib.h>

/* from the packet man page so that we compile under libc5 or glibc */
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <features.h>    /* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>     /* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>   /* The L2 protocols */
#endif

#define USERLAND_SESSIONS	1

int pppoe_session::HardOutput(CPppPacket *pkt)
{
	pkt->Push16(pkt->GetLength() /* length */);
	pkt->Push16(session_id);
	pkt->Push8(0x00 /* code */);
	pkt->Push8(0x11 /* ver type */);
	pkt->Push16(ETH_P_PPPoE_SESN /* ether type */);
	pkt->Push8(parent->our_addr[5]);
	pkt->Push8(parent->our_addr[4]);
	pkt->Push8(parent->our_addr[3]);
	pkt->Push8(parent->our_addr[2]);
	pkt->Push8(parent->our_addr[1]);
	pkt->Push8(parent->our_addr[0]);

	pkt->Push8(peer_mac[5]);
	pkt->Push8(peer_mac[4]);
	pkt->Push8(peer_mac[3]);
	pkt->Push8(peer_mac[2]);
	pkt->Push8(peer_mac[1]);
	pkt->Push8(peer_mac[0]);

	int len = pkt->GetLength();
	int ret = sendto(parent->pppoed_disc_fd, pkt->m_start, len, 0,
			 &parent->sa, sizeof parent->sa);
	if (len != ret)
		perror("pppoe_session::HardOutput: write(pppoed_disc_fd)");

	return ret;
}

void pppoed::send_padt(u16 session_id, u8 *our_addr, u8 *peer_addr)
{
	/* make a PADT */
	CPppPacket padt;

	padt.Push16(padt.GetLength() /* length */);
	padt.Push16(session_id);
	padt.Push8(PPPoE_CODE_PADT /* code */);
	padt.Push8(0x11 /* ver type */);
	padt.Push16(ETH_P_PPPoE_SESN /* ether type */);
	padt.Push8(our_addr[5]);
	padt.Push8(our_addr[4]);
	padt.Push8(our_addr[3]);
	padt.Push8(our_addr[2]);
	padt.Push8(our_addr[1]);
	padt.Push8(our_addr[0]);

	padt.Push8(peer_addr[5]);
	padt.Push8(peer_addr[4]);
	padt.Push8(peer_addr[3]);
	padt.Push8(peer_addr[2]);
	padt.Push8(peer_addr[1]);
	padt.Push8(peer_addr[0]);

	int len = padt.GetLength();
	int ret = sendto(this->pppoed_disc_fd, padt.m_start, len, 0,
			 &sa, sizeof sa);
	if (len != ret)
		perror("pppoe_session::HardOutput: write(pppoed_disc_fd)");
	/* sent PADT */
}

void pppoe_session::HardHangup(void)
{
	ch.state = CS_DISCONNECTING;

	parent->send_padt(session_id, parent->our_addr, peer_mac);

	ConnectComplete(0x400);
	delete this;
}

void pppoe_session::pppoe_input(u8 *buf, int len)
{
	struct pppoehdr *h = (struct pppoehdr *)buf;
	buf += sizeof *h;
	len -= sizeof *h;

	if (h->src_addr[0] != peer_mac[0] || 
	    h->src_addr[1] != peer_mac[1] || 
	    h->src_addr[2] != peer_mac[2] || 
	    h->src_addr[3] != peer_mac[3] || 
	    h->src_addr[4] != peer_mac[4] || 
	    h->src_addr[5] != peer_mac[5]) {
		if (h->code != PPPoE_CODE_PADT)
			parent->send_padt(ntohs(h->sesn_id), h->dst_addr, h->src_addr);
		return;
	}

	if (h->ether_type == htons(ETH_P_PPPoE_SESN) && h->code == PPPoE_CODE_DATA) {
		CLink *link = this;
		u16 h_len = ntohs(h->length);
		if (h_len < len)
			len = h_len;
		link->Input(buf, len);
		return;
	}

	if (h->code == PPPoE_CODE_PADT) {
		ch.state = CS_DISCONNECTING;
		ConnectComplete(0x400);
		delete this;
	}
}

pppoed::pppoed(const char *dev_name, int disc_fd, char *addr)
{
	memset(&sa, 0, sizeof sa);
	sa.sa_family = 0;
	strncpy(pppoed_dev_name, dev_name, sizeof pppoed_dev_name);
	strcpy(sa.sa_data, pppoed_dev_name);

	pppoed_disc_fd = disc_fd;
	pppoed_sesn_fd = -1;
	memcpy(our_addr, addr, sizeof our_addr);
	last_session_id = 0;
	memset(&sessions, 0, sizeof sessions);

	SelectAddEvent(disc_fd, SEL_READ);
}

void pppoed::send_pado(struct pppoehdr *h)
{
#define mktag(type, tagdata, taglen)		\
	do {					\
		*(u16 *)buf = htons(type);	\
		buf += 2;			\
		*(u16 *)buf = htons(taglen);	\
		buf += 2;			\
		memcpy(buf, tagdata, taglen);	\
		buf += taglen;			\
	} while (0)

	unsigned char tmp[1500];
	unsigned char *buf = tmp;
	struct pppoehdr *reply = (struct pppoehdr *)buf;

	buf += sizeof(*reply);
	memcpy(reply->dst_addr, h->src_addr, 6);
	memcpy(reply->src_addr, our_addr, 6);
	reply->ether_type = htons(ETH_P_PPPoE_DISC);
	reply->vertype = 0x11;
	reply->code = PPPoE_CODE_PADO;
	reply->sesn_id = 0x0000;
	reply->length = 0;

	mktag(PPPoE_TAG_AC_NAME, "Babylon", 7);
	mktag(PPPoE_TAG_SERVICE_NAME, "internet", 8);

	reply->length = htons((buf - tmp) - sizeof(*reply));

	if ((buf - tmp) != sendto(pppoed_disc_fd, tmp, buf - tmp, 0, &sa, sizeof sa))
		perror("write(pppoed_disc_fd)");
}

pppoe_session::pppoe_session(u8 *mac, u16 id, pppoed *pppoed)
{
	memcpy(peer_mac, mac, sizeof peer_mac);
	session_id = id;
	parent = pppoed;

	memset(&ch, 0, sizeof ch);
	char tmp[16];
	sprintf(tmp, "pppoe%u", session_id);
	ch.device_name = strdup(tmp);
	ch.dev_class = "pppoe";
	ch.link = this;
	ch.link->m_hard_mtu = 1492;
	ch.link->m_hard_mru = 1492;
	m_channel = &ch;
}

pppoe_session::~pppoe_session()
{
	if (parent) {
		parent->remove_session(this, session_id);
		parent = NULL;
	}
	remove_disc_chans();
	remove_multihop_link();
	free((void *)ch.device_name);
	ch.device_name = NULL;
}

u16 pppoed::alloc_sesn_id(unsigned char *mac_addr)
{
	int i;
	for (i=0; i<65535; i++) {
		last_session_id ++;
		if (!last_session_id)
			last_session_id++;

		if (!sessions[last_session_id]) {
			pppoe_session *session = new pppoe_session(mac_addr, last_session_id, this);
			sessions[last_session_id] = session;
			session->Open();
			session->m_channel->state = CS_CONNECTED;
			session->Up();
			return last_session_id;
		}
	}
	return 0;
}

void pppoed::send_pads(struct pppoehdr *h)
{
	unsigned char tmp[1500];
	unsigned char *buf = tmp;
	struct pppoehdr *reply = (struct pppoehdr *)buf;

	buf += sizeof(*reply);
	memcpy(reply->dst_addr, h->src_addr, 6);
	memcpy(reply->src_addr, our_addr, 6);
	reply->ether_type = htons(ETH_P_PPPoE_DISC);
	reply->vertype = 0x11;
	reply->code = PPPoE_CODE_PADS;
	reply->sesn_id = htons(alloc_sesn_id(h->src_addr));
	reply->length = 0;

	mktag(PPPoE_TAG_AC_NAME, "Babylon", 7);
	mktag(PPPoE_TAG_SERVICE_NAME, "internet", 8);

	reply->length = htons((buf - tmp) - sizeof(*reply));

	if ((buf - tmp) != sendto(pppoed_disc_fd, tmp, buf - tmp, 0, &sa, sizeof sa))
		perror("write(pppoed_disc_fd)");
}

void pppoed::SelectEvent(int fd, SelectEventType event)
{
	if (!(event & SEL_READ))
		return;

	unsigned char buf[4096];
	int len = read(fd, buf, sizeof buf);

	if (len < 0)
		return;

	struct pppoehdr *h = (struct pppoehdr *)buf;

	if (h->vertype != 0x11) {
		fprintf(stderr, "pppoe: wrong version 0x%02x\n", h->vertype);
		return;
	}

#if 0
	fprintf(stderr, "from: %02x:%02x:%02x:%02x:%02x:%02x "
		"to: %02x:%02x:%02x:%02x:%02x:%02x\n",
		h->src_addr[0], h->src_addr[1], h->src_addr[2],
		h->src_addr[3], h->src_addr[4], h->src_addr[5],
		h->dst_addr[0], h->dst_addr[1], h->dst_addr[2],
		h->dst_addr[3], h->dst_addr[4], h->dst_addr[5]);
	fprintf(stderr, "sesn(0x%04x) pppoe length = %d (read = %d)\n",
		ntohs(h->sesn_id), ntohs(h->length), len);
#endif

	switch(h->code) {
	case PPPoE_CODE_PADT:
		fprintf(stderr, "got PADT\n");
	case 0:
		if (h->sesn_id) {
			pppoe_session *session =  sessions[ntohs(h->sesn_id)];
			if (session) {
				session->pppoe_input(buf, len);
				return;
			}
			if (h->code != PPPoE_CODE_PADT)
				send_padt(ntohs(h->sesn_id), h->dst_addr, h->src_addr);
			return;
		}
		break;

	case PPPoE_CODE_PADI:
		fprintf(stderr, "got PADI\n");
		if (h->sesn_id != 0) {
			fprintf(stderr, "dropping PADI\n");
			return;
		}
		send_pado(h);
		break;
	case PPPoE_CODE_PADO:
		fprintf(stderr, "got PADO\n");
		break;
	case PPPoE_CODE_PADS:
		fprintf(stderr, "got PADS\n");
		break;
	case PPPoE_CODE_PADR:
		fprintf(stderr, "got PADR\n");
		if (h->dst_addr[0] == 0xff && h->dst_addr[1] == 0xff &&
		    h->dst_addr[2] == 0xff && h->dst_addr[3] == 0xff &&
		    h->dst_addr[4] == 0xff && h->dst_addr[5] == 0xff) {
			fprintf(stderr, "dropping PADR sent to broadcast\n");
			return;
		}
		send_pads(h);
		break;
	}
	return;
}

void pppoed::setup_sesn_sk(int sk)
{
	pppoed_sesn_fd = sk;
	SelectAddEvent(sk, SEL_READ);
}

void pppoed::remove_session(pppoe_session *session, u16 idx)
{
	if (sessions[idx] != session)
		*(char *)0 = 0;
	sessions[idx] = NULL;
}

static int __bpppoe_listen(ctrlfd_t *cfd, const char *devname, u16 type, char *our_addr)
{
	cfd->printf("listen '%s'\n", devname);

	int sk = socket(PF_INET, SOCK_PACKET, type);
	if (sk < 0) {
		cfd->perror("socket");
		cfd->done(1);
		return -1;
	}

	struct ifreq ifreq;

	memset(&ifreq, 0, sizeof(ifreq));
	strncpy(ifreq.ifr_name, devname, sizeof(ifreq.ifr_name));
	if (ioctl(sk, SIOCGIFHWADDR, &ifreq) < 0) {
		cfd->perror("SIOCGIFHWADDR:");
		cfd->done(1);
		close(sk);
		return -1;
	}

	if (our_addr)
		memcpy(our_addr, &ifreq.ifr_ifru.ifru_hwaddr.sa_data, 6);

	struct sockaddr sa;
	sa.sa_family = 0;
	strncpy(sa.sa_data, devname, sizeof(sa.sa_data));

	if (bind(sk, &sa, sizeof(sa)) < 0) {
		cfd->perror("bind");
		cfd->done(1);
		close(sk);
		return -1;
	}

	return sk;
}

static void bpppoe_listen(ctrlfd_t *cfd, const char *devname)
{
	char our_addr[6];
	int sk = __bpppoe_listen(cfd, devname, htons(ETH_P_PPPoE_DISC), our_addr);
	if (sk < 0)
		return;

	pppoed *daemon = new pppoed(devname, sk, our_addr);

#ifdef USERLAND_SESSIONS
	int sesn_sk = __bpppoe_listen(cfd, devname, htons(ETH_P_PPPoE_SESN), NULL);
	if (sesn_sk < 0) {
		delete daemon;
		return;
	}
	daemon->setup_sesn_sk(sesn_sk);
#endif

	cfd->printf("okay\n");
	cfd->done(0);
}

void do_bpppoe(ctrlfd_t *cfd, const char *str)
{
	if (!strncmp("listen ", str, 7))
		return bpppoe_listen(cfd, str+7);

	cfd->printf("usage:\n");
	cfd->printf("	bpppoe listen <dev>\n");
	cfd->done(2);
}

