#include "selectops.h"
#include "link.h"

#include <sys/types.h>
#include <sys/socket.h>

#define ETH_P_PPPoE_DISC	0x8863
#define ETH_P_PPPoE_SESN	0x8864

#define PPPoE_CODE_DATA		0x00
#define PPPoE_CODE_PADI		0x09
#define PPPoE_CODE_PADO		0x07
#define PPPoE_CODE_PADR		0x19
#define PPPoE_CODE_PADS		0x65
#define PPPoE_CODE_PADT		0xa7

#define PPPoE_TAG_END_OF_LIST           0x0000
#define PPPoE_TAG_SERVICE_NAME          0x0101
#define PPPoE_TAG_AC_NAME               0x0102
#define PPPoE_TAG_HOST_UNIQ             0x0103
#define PPPoE_TAG_AC_COOKIE             0x0104
#define PPPoE_TAG_VENDOR_SPECIFIC       0x0105
#define PPPoE_TAG_RELAY_SESSION_ID      0x0110
#define PPPoE_TAG_SERVICE_NAME_ERROR    0x0201
#define PPPoE_TAG_AC_SYSTEM_ERROR       0x0202
#define PPPoE_TAG_GENERIC_ERROR         0x0203

struct pppoehdr {
	unsigned char	dst_addr[6];
	unsigned char	src_addr[6];
	unsigned short	ether_type;
	unsigned char	vertype;
	unsigned char	code;
	unsigned short	sesn_id;
	unsigned short	length;
} __attribute__((packed));

class pppoed;

class pppoe_session : public CLink {
	u8	peer_mac[6];
	u16	session_id;
	pppoed	*parent;


public:
	/* We need to default to a 1492 MTU for PPPoE sessions, otherwise 
	 * LCP won't converge with a DLink WBR-2310 router.
	 */
	virtual unsigned DefaultMRU(void) {
		return 1492;
	}

	channel_t ch;
	pppoe_session(u8 *mac, u16 id, pppoed *pppoed);
	pppoe_session(channel_t *ch);
	~pppoe_session();
	void pppoe_input(u8 *buf, int len);

	int HardOutput(CPppPacket *pkt);
	void HardHangup(void);
};

class pppoed : public SelectEventHandler {
	friend int pppoe_session::HardOutput(CPppPacket *pkt);
	friend void pppoe_session::HardHangup(void);

	int	pppoed_disc_fd;
	int	pppoed_sesn_fd;
	char	pppoed_dev_name[16];
	u8	our_addr[6];
	u16		last_session_id;

	pppoe_session	*sessions[65536];

	u16 alloc_sesn_id(unsigned char *mac_addr);
	void send_pado(struct pppoehdr *h);
	void send_pads(struct pppoehdr *h);

public:
	void send_padt(u16 session_id, u8 *our_addr, u8 *peer_addr);
	struct sockaddr sa;
	pppoed(const char *dev_name, int disc_fd, char *addr);
	void setup_sesn_sk(int sk);
	void SelectEvent(int fd, SelectEventType event);
	void remove_session(pppoe_session *s, u16 idx);
};

